/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
/*      ###           Implementacia interpretu jazyka IFJ09          ###      */  
/* ########################################################################## */
/*  Akademicky rok: 2009/2010 ## Predmet: IFJ ## Datum odovzdania : 13.12 09  */
/* |------------------------------------------------------------------------| */
/* |        Team              |   Login   |       Nazov a popis modulu      | */ 
/* |------------------------------------------------------------------------| */
/* | Martin Haviar - veduci   | xhavia01  |                                 | */
/* | Jozef  Henzl             | xhenzl04  |                                 | */
/* | Martin Herich            | xheric00  |                                 | */
/* | Maros  Barjak            | xbarja00  |                                 | */
/* \------------------------------------------------------------------------/ */
                                                                 
/*
 *	hashovacia tabulka ako tabulka symbolov 
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>

#include "hash.h"

/*
 *   Vytvorenie a inicializacia tabulky
 */
t_htable* htable_init(int size)
{
    t_htable *result = malloc(sizeof(t_htable));
    if (result==NULL) return NULL;

    result->htable_size=size;

    result->p = (htable_item **) malloc(size * sizeof(htable_item *));
    if (result->p == NULL) return NULL;

    for (int i=0; i<size; i++)
    {
        result->p[i] = malloc(sizeof(htable_item));
        if (result->p[i]==NULL) return NULL;

        result->p[i]->next=NULL;
    }

    return result;
}


/*
 *   Uvolnenie vsetkych poloziek tabulky z pamate (vyprazdnenie)
 */
void htable_clear(t_htable* htable)
{
    for (int i=0; i<htable->htable_size; i++)
 	   {
        htable_item *del = htable->p[i]->next; // odkaz na prvy zaznam / NULL
        while (del!=NULL)
        {
            if (del->type==T_STRING) strFree(&del->data._string);
            htable_item *tmp = del;
            del = del->next;
            strFree(&tmp->name);
            free(tmp);
        }
    }
}


/*
 *   Vyprazdnenie a uvolnenie tabulky z pamate
 */
void htable_dispose(t_htable* htable)
{
    assert(htable);

	htable_clear(htable);	// vyprazdnenie tabulky

    for (int i=0; i< htable->htable_size; i++ ) {
        free(htable->p[i]);
	}

    free(htable->p);
    free(htable);
    htable=NULL;
}


/*
 *   Rozptylovacia funkcia pre retazce
 */
unsigned int hash_function(const char *str, unsigned htable_size)
{
    unsigned int h=0;
    unsigned char *p;

    for (p=(unsigned char*)str; *p!='\0'; p++)
        h = 31*h + *p;

    return h % htable_size;
}



/*
 *   Vypis zaznamov z hash table (debugovacie vypisy)
 */
void _htable_write(t_htable *htable)
{

    for (int i=0; i<htable->htable_size; i++)
    {
        htable_item *tmp = htable->p[i]->next; // odkaz na prvy zaznam / NULL
        while (tmp!=NULL)
        {
			switch (tmp->type)
			{
			case T_INTEGER:
				fprintf(stderr,"[%d]\t%s\tint\t%d\n",i,tmp->name.str,tmp->data.num_int);	
			break;
			case T_DOUBLE:
				fprintf(stderr,"[%d]\t%s\tdouble\t%f\n",i,tmp->name.str,tmp->data.num_double);	
			break;
			case T_STRING:
				fprintf(stderr,"[%d]\t%s\tstring\t%s\n",i,tmp->name.str,tmp->data._string.str);	
			break;
			}
									
            tmp = tmp->next;
        }
    }
fprintf(stderr,"--- koniec tabulky ---\n");
}



/*
 *   Funkcia vyhlada v tabulke t zaznam podla name,
 *	 ak najde tak vrati 1 a p=odkaz na najdeny zaznam,
 *   ak nenajde tak prida novy zaznam s parametrami name, type, tmp a vrati 0
 *	 ak dojde ku chybe pri alokacii vrati -1   
 */
int htable_lookup(t_htable *t, htable_item **pp, char *name, int tmp)
{

	assert(t);

    // ziska index do hash table podla rozptylovacej funkcie
    unsigned int index = hash_function(name, t->htable_size);
	htable_item *last = t->p[index];	// kam napojit novy prvok ak sa nenajde name
    htable_item *p;

    // prehladava zaznamy pod dannym indexom
    for (htable_item *i =t->p[index]->next ; i!=NULL ; i = i->next  )
    {
        if (strcmp(i->name.str,name)==0)
			{
				// nasiel sa zaznam s hladanym name
				*pp = i;
            	return HT_FOUND;	// zaznam uz existoval, p nan ukazuje
			}
		last = i;	
    }

    // name sa nenasiel - vytvor zaznam a vrat nanho ukazatel
    p = calloc(1,sizeof(htable_item));
    if (p==NULL) return HT_ERROR;	// chyba pri alokacii pamate

    //nastav referenciu na novy zaznam
    last->next = p;

    // nastavenie mena podla parametru name
    if (strInit(&p->name) != STR_SUCCESS) return HT_ERROR;
    if (strConcat(&p->name,name)!=STR_SUCCESS) return HT_ERROR;

    p->type = T_INTEGER;    // defaultne int
    p->next = NULL;         // pridava sa ako posledny zaznam

    // odkazeme pp na novy prvok
    *pp = p;

    return HT_INSERTED;	// bol pridany novy zaznam

}

#if 0
 // POUZITIE
int main()
{

	t_htable *a = htable_init(10);
	htable_item *p;

	//  union sa sprava ako miesto v pamati, kt moze nadobudat viac typov hodnot
	//	ale vzdy len jeden typ v danom okamihu

	htable_lookup(a, &p, "prvy",  0 );
    p->type = T_INTEGER;
    p->data.num_int = 100;

	htable_lookup(a, &p, "druhy", 0 );
    p->type = T_DOUBLE;
    p->data.num_double = 3.14987654;

	htable_lookup(a, &p, "treti", 0 );
    p->type = T_STRING;
    p->data.str = "holub";

	htable_write(a);

	htable_dispose(a);
	return(0);
}
#endif
