/**
* @brief 
* @file inst_lst.c
* @author J.H. RADAS 2009
* @date 2009-11-21
*/

/* system includes */
#include <stdlib.h>
#include <assert.h>
#include <string.h>

/* local includes */
#include "inst_lst.h"
#include "error.h"

#define SEG_SIZE	64

ilist_t *ilist_new()
{
	ilist_t *new;
	new = calloc(1, sizeof(ilist_t));
	
	return new;
}

il_addr_t ilist_last_addr(ilist_t *ilist)
{
	return ilist->_a_max - 1;
}


instruction_t *ilist_addi_end(ilist_t *ilist)
{
	instruction_t *inst;
	instruction_t **il_new;
	il_addr_t seg, off;

	/* je potreba alokovat novou pamet */
	if(ilist->_a_max == ilist->_a_size) {
		il_new = realloc(ilist->instr, sizeof(instruction_t**) * (ilist->_segs + 1));
		if(il_new == NULL) return NULL;
		ilist->instr = il_new;

		inst = calloc(1, sizeof(instruction_t) * SEG_SIZE);
		ilist->instr[ilist->_segs] = inst;

		ilist->_a_size += SEG_SIZE;
		ilist->_segs++;
	}
	seg = ilist->_a_max / SEG_SIZE;
	off = ilist->_a_max % SEG_SIZE;
	
	inst = &ilist->instr[seg][off];
	ilist->_a_max++;
	memset(inst, 0, sizeof(instruction_t));
	return inst;
}

void ilist_free(ilist_t *ilist)
{
	int i;
	assert(ilist != NULL);

	for(i = 0;i < ilist->_segs; i++) {
		free(ilist->instr[i]);
	}
	free(ilist->instr);

	free(ilist);
}

int ilist_goto(ilist_t *ilist, il_addr_t addr)
{
	assert(ilist != NULL);
	
	/* kontrola jestli se nesnazime skocit mimo nas adresni prostor */
	if (addr > ilist->_a_max) {
		return -1;
	}
	
	ilist->_a_cur = addr;
	return 0;
}


instruction_t *ilist_getnext_inst(ilist_t *ilist)
{
	il_addr_t seg, off;
	if(ilist->_a_cur == ilist->_a_max) return NULL;	/* end of program */

	seg = ilist->_a_cur / SEG_SIZE;
	off = ilist->_a_cur % SEG_SIZE;
	ilist->_a_cur++;

	return &ilist->instr[seg][off];
}

#if 0
#include "interpret.h"
int main(int argc, char **argv)
{
	ilist_t *ilist;
	instruction_t *inst;

	t_htable *htable;
	htable_item *ht_i;

	ilist = ilist_new();
	htable = htable_init(256);
	htable_lookup(htable, &ht_i, "a", 0);
	ht_i->type = T_INTEGER;

	inst = ilist_addi_end(ilist);
	inst->type = INST_READ;
	inst->i1 = ht_i;

	inst = ilist_addi_end(ilist);
	inst->type = INST_WRITE;
	inst->i1 = ht_i;

	interpret_run(ilist);

	htable_dispose(htable);
	ilist_free(ilist);


}
#endif
