/**
* @brief 
* @file interpret.c
* @author J.H. RADAS 2009
* @date 2009-11-21
*/

/* system includes */
#include <stdio.h>

/* local includes */
#include "interpret.h"
#include "inst_lst.h"
#include "error.h"
#include "str.h"

/* function prototypes */
static int __write(htable_item *i1);
static int __read(htable_item *i1);
static int __add(htable_item *i1, htable_item *i2, htable_item *i3);
static int __sub(htable_item *i1, htable_item *i2, htable_item *i3);
static int __mul(htable_item *i1, htable_item *i2, htable_item *i3);
static int __div(htable_item *i1, htable_item *i2, htable_item *i3);
static int __mov(htable_item *i1, htable_item *i2, htable_item *i3);
static int __cmp_less(htable_item *i1, htable_item *i2, htable_item *i3);
static int __cmp_great(htable_item *i1, htable_item *i2, htable_item *i3);
static int __cmp_lesseq(htable_item *i1, htable_item *i2, htable_item *i3);
static int __cmp_greateq(htable_item *i1, htable_item *i2, htable_item *i3);
static int __cmp_equal(htable_item *i1, htable_item *i2, htable_item *i3);
static int __cmp_notequal(htable_item *i1, htable_item *i2, htable_item *i3);
static int __jmp(ilist_t *ilist, htable_item *i1);
static int __jz(ilist_t *ilist, htable_item *i1, htable_item *i2);
static int __find(htable_item *i1, htable_item *i2, htable_item *i3);
static int __sort(htable_item *i1, htable_item *i3);
static int __clr();

int interpret_run(ilist_t *ilist)
{
	il_addr_t caddr;	/**< current address */
	instruction_t *inst=NULL;
	int err=0;

	caddr = 0;	
	ilist_goto(ilist, 0);	
	while(1) {
		if(err < 0) return err;
		inst = ilist_getnext_inst(ilist);
		if(inst == NULL) break;
		switch(inst->type) {
		case INST_READ:
			err = __read(inst->i1);
		break;
		case INST_WRITE:
			err = __write(inst->i1);
		break;
		case INST_ADD:
			err = __add(inst->i1, inst->i2, inst->i3);
		break;
		case INST_SUB:
			err = __sub(inst->i1, inst->i2, inst->i3);
		break;
		case INST_MUL:
			err = __mul(inst->i1, inst->i2, inst->i3);
		break;
		case INST_DIV:
			err = __div(inst->i1, inst->i2, inst->i3);
		break;
		case INST_MOV:
			err = __mov(inst->i1, NULL,  inst->i3);
		break;
		case INST_CMP_LESS:
			err = __cmp_less(inst->i1, inst->i2, inst->i3);
		break;
		case INST_CMP_GREAT:
			err = __cmp_great(inst->i1, inst->i2, inst->i3);
		break;
		case INST_CMP_GREATEQ:
			err = __cmp_greateq(inst->i1, inst->i2, inst->i3);
		break;
		case INST_CMP_LESSEQ:
			err = __cmp_lesseq(inst->i1, inst->i2, inst->i3);
		break;
		case INST_CMP_EQUAL:
			err = __cmp_equal(inst->i1, inst->i2, inst->i3);
		break;
		case INST_CMP_NOTEQUAL:
			err = __cmp_notequal(inst->i1, inst->i2, inst->i3);
		break;
		case INST_JMP:
			err = __jmp(ilist, inst->i3);
		break;
		case INST_JZ:
			err = __jz(ilist, inst->i1, inst->i3);;
		break;
		case INST_FIND:
			err = __find(inst->i1, inst->i2, inst->i3);
		break;
		case INST_SORT:
			err = __sort(inst->i1, inst->i3);
		break;
		case INST_CLR:
			err = __clr();
		break;
		case INST_LABEL:
			continue;
		default:	
			print_log("unknown instruction: %d", inst->type);
			return -1;
		}		
	}
	return err;
}

/**
* nepodmineny skok na adresu danou i1
*/
static int __jmp(ilist_t *ilist, htable_item *i3)
{
	if(i3->type != T_INTEGER) {
		return -1;
	}
	ilist_goto(ilist, i3->data.num_int);
	return 0;
}

/**
* skok na adresu danou hodnotou instrukce 3 pokud je i1 nulova
*/
static int __jz(ilist_t *ilist, htable_item *i1, htable_item *i3)
{
	if((i1->type != T_INTEGER) || (i3->type != T_INTEGER)) {    
		return -1;
	}
	if(i1->data.num_int == 0)
		ilist_goto(ilist, i3->data.num_int);
	return 0;
}

static int __read(htable_item *i1)
{
	int in;
	switch(i1->type) {
		case T_INTEGER:
			in = scanf("%d", &i1->data.num_int);
		break;
		case T_DOUBLE:
			in = scanf("%lf", &i1->data.num_double);
		break;
		case T_STRING:
			strClear(&i1->data._string);
			while((in = getchar())) {
				if(in == '\n') {
					ungetc(in, stdin);
					return 0;
				}
				else if(in == EOF) return -1;
				strAddChar(&i1->data._string, in);
			}
		break;
		default: return -1;
	}
	if(in == 0) {
		return -1;
	}
	return 0;
}

static int __write(htable_item *i1)
{
	switch(i1->type) {
		case T_INTEGER:
			printf("%d", i1->data.num_int);
		break;
		case T_DOUBLE:
			printf("%g", i1->data.num_double);
		break;
		case T_STRING:
			printf("%s", i1->data._string.str);
		break;
		default:
			return -1;
	}
	return 0;
}

static int __add(htable_item *i1, htable_item *i2, htable_item *i3)
{
	switch(i1->type) {
		case T_INTEGER:
			i3->type = T_INTEGER;
			if(i2->type == T_INTEGER) 
				i3->data.num_int = i1->data.num_int + i2->data.num_int;
			else if(i2->type == T_DOUBLE) {
				i3->type = T_DOUBLE;
				i3->data.num_double = i1->data.num_int + i2->data.num_double;
			}
		break;
		case T_DOUBLE:
			i3->type = T_DOUBLE;
			if(i2->type == T_INTEGER) 
				i3->data.num_double = i1->data.num_double + i2->data.num_int;
			else if(i2->type == T_DOUBLE) 
				i3->data.num_double = i1->data.num_double + i2->data.num_double;
		break;
		case T_STRING:
			strClear(&i3->data._string);
			strDup(&i3->data._string, &i1->data._string);
			strConcat(&i3->data._string, i2->data._string.str);
		break;
		default:
			return -1;
	}
	return 0;
}
static int __sub(htable_item *i1, htable_item *i2, htable_item *i3)
{
	i3->type = T_INTEGER;
	switch(i1->type) {
		case T_INTEGER:
			if(i2->type == T_INTEGER) 
				i3->data.num_int = i1->data.num_int - i2->data.num_int;
			else if(i2->type == T_DOUBLE) {
				i3->type = T_DOUBLE;
				i3->data.num_double = i1->data.num_int - i2->data.num_double;
			}
		break;
		case T_DOUBLE:
			i3->type = T_DOUBLE;
			if(i2->type == T_INTEGER) 
				i3->data.num_double = i1->data.num_double - i2->data.num_int;
			else if(i2->type == T_DOUBLE) 
				i3->data.num_double = i1->data.num_double - i2->data.num_double;
		break;
		default:
			return -1;
	}
	return 0;
}

static int __mul(htable_item *i1, htable_item *i2, htable_item *i3)
{
	i3->type = T_INTEGER;
	switch(i1->type) {
		case T_INTEGER:
			if(i2->type == T_INTEGER) 
				i3->data.num_int = i1->data.num_int * i2->data.num_int;
			else if(i2->type == T_DOUBLE) {
				i3->type = T_DOUBLE;
				i3->data.num_double = i1->data.num_int * i2->data.num_double;
			}
		break;
		case T_DOUBLE:
			i3->type = T_DOUBLE;
			if(i2->type == T_INTEGER) 
				i3->data.num_double = i1->data.num_double * i2->data.num_int;
			else if(i2->type == T_DOUBLE) 
				i3->data.num_double = i1->data.num_double * i2->data.num_double;
		break;
		default:
			return -1;
	}
	return 0;
}

static int __div(htable_item *i1, htable_item *i2, htable_item *i3)
{
	i3->type = T_INTEGER;
	switch(i1->type) {
		case T_INTEGER:
			if(i1->type == T_INTEGER)  {
				if(i2->data.num_int == 0) return -1;
				i3->data.num_int = i1->data.num_int / i2->data.num_int;
			} else if(i1->type == T_DOUBLE)
				return -1;
		break;
		case T_DOUBLE:
		default:
			return -1;
	}
	return 0;
}

static int __cmp_less(htable_item *i1, htable_item *i2, htable_item *i3)
{
	if(i1->type != i2->type) return -1;
	i3->type = T_INTEGER;
	switch(i1->type) {
		case T_INTEGER:
			i3->data.num_int = (i1->data.num_int < i2->data.num_int) ? 1 : 0;
		break;
		case T_DOUBLE:
			i3->data.num_int = (i1->data.num_double < i2->data.num_double) ? 1 : 0;
		break;
		case T_STRING:
			i3->data.num_int = (strCmpString(&i1->data._string, &i2->data._string) < 0) ? 1 : 0;
		break;
		default:
			return -1;
	}
	return 0;
}

static int __cmp_lesseq(htable_item *i1, htable_item *i2, htable_item *i3)
{
	if(i1->type != i2->type) return -1;
	i3->type = T_INTEGER;
	switch(i1->type) {
		case T_INTEGER:
			i3->data.num_int = (i1->data.num_int <= i2->data.num_int) ? 1 : 0;
		break;
		case T_DOUBLE:
			i3->data.num_int = (i1->data.num_double <= i2->data.num_double) ? 1 : 0;
		break;
		case T_STRING:
			i3->data.num_int = (strCmpString(&i1->data._string, &i2->data._string) <= 0) ? 1 : 0;
		break;
		default:
			return -1;
	}
	return 0;
}

static int __cmp_great(htable_item *i1, htable_item *i2, htable_item *i3)
{
	if(i1->type != i2->type) return -1;
	i3->type = T_INTEGER;
	switch(i2->type) {
		case T_INTEGER:
			i3->data.num_int = (i1->data.num_int > i2->data.num_int) ? 1 : 0;
		break;
		case T_DOUBLE:
			i3->data.num_int = (i1->data.num_double > i2->data.num_double) ? 1 : 0;
		break;
		case T_STRING:
			i3->data.num_int = (strCmpString(&i1->data._string, &i2->data._string) > 0) ? 1 : 0;
		break;
		default:
			return -1;
	}
	return 0;
}

static int __cmp_greateq(htable_item *i1, htable_item *i2, htable_item *i3)
{
	if(i1->type != i2->type) return -1;
	i3->type = T_INTEGER;
	switch(i2->type) {
		case T_INTEGER:
			i3->data.num_int = (i1->data.num_int >= i2->data.num_int) ? 1 : 0;
		break;
		case T_DOUBLE:
			i3->data.num_int = (i1->data.num_double >= i2->data.num_double) ? 1 : 0;
		break;
		case T_STRING:
			i3->data.num_int = (strCmpString(&i1->data._string, &i2->data._string) >= 0) ? 1 : 0;
		break;
		default:
			return -1;
	}
	return 0;
}

static int __cmp_equal(htable_item *i1, htable_item *i2, htable_item *i3)
{
	if(i1->type != i2->type) return -1;
	i3->type = T_INTEGER;
	switch(i1->type) {
		case T_INTEGER:
			i3->data.num_int = (i1->data.num_int == i2->data.num_int) ? 1 : 0;
		break;
		case T_DOUBLE:
			i3->data.num_int = (i1->data.num_double == i2->data.num_double) ? 1 : 0;
		break;
		case T_STRING:
			i3->data.num_int = (strCmpString(&i1->data._string, &i2->data._string) == 0) ? 1 : 0;
		break;
		default:
			return -1;
	}
	return 0;
}

static int __cmp_notequal(htable_item *i1, htable_item *i2, htable_item *i3)
{
	if(i1->type != i2->type) return -1;
	i3->type = T_INTEGER;
	switch(i1->type) {
		case T_INTEGER:
			i3->data.num_int = (i1->data.num_int != i2->data.num_int) ? 1 : 0;
		break;
		case T_DOUBLE:
			i3->data.num_int = (i1->data.num_double != i2->data.num_double) ? 1 : 0;
		break;
		case T_STRING:
			i3->data.num_int = (strCmpString(&i1->data._string, &i2->data._string) != 0) ? 1 : 0;
		break;
		default:
			return -1;
	}
	return 0;
}

/**
* prirazeni hodnoty
* parametr i2 je NULL
*/
static int __mov(htable_item *i1, htable_item *i2, htable_item *i3)
{
	switch(i3->type) {
		case T_INTEGER:
			if (i1->type == T_INTEGER) 
				i3->data.num_int = i1->data.num_int; 
			else if(i1->type == T_DOUBLE)
				i3->data.num_int = (int)i1->data.num_double;
		    break;
		case T_DOUBLE:
			if (i1->type == T_INTEGER)
				i3->data.num_double = (double)i1->data.num_int;
			else if(i1->type == T_DOUBLE)
				i3->data.num_double = i1->data.num_double;
		    break;		    
		case T_STRING:
			if (i1->type == T_STRING)
				strDup(&i3->data._string, &i1->data._string);
		    break;
		    
		default:
			return -1;
	}
	return 0;
}

static int __find(htable_item *i1, htable_item *i2, htable_item *i3)
{
	int err;

	if((i1->type != T_STRING) && (i2->type != T_STRING)) {
		return -1;
	}
	
	err = find(i1->data._string.str, i2->data._string.str);
	i3->type = T_INTEGER;
	i3->data.num_int = (err < 0) ? 0 : err+1;
	return 0;
}


static int __sort(htable_item *i1, htable_item *i3)
{
	int err;

	if(i1->type != T_STRING) {
		return -1;
	}

	i3->type = T_STRING;
	if (i1 != i3) {
		strDup(&i3->data._string, &i1->data._string);
	}
	sort(i3->data._string.str);

	return err;
}

/*
* nacitej vstup az do konce radku, pokud narazis na EOF tak chyba
*/
static int __clr()
{
	int in;
	while((in = getchar()) != '\n') {
		if (in == EOF)	return -1;
	}
	return 0;
}
