/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
/*      ###           Implementacia interpretu jazyka IFJ09          ###      */  
/* ########################################################################## */
/*  Akademicky rok: 2009/2010 ## Predmet: IFJ ## Datum odovzdania : 13.12 09  */
/* |------------------------------------------------------------------------| */
/* |        Team              |   Login   |       Nazov a popis modulu      | */ 
/* |------------------------------------------------------------------------| */
/* | Martin Haviar - veduci   | xhavia01  |                                 | */
/* | Jozef  Henzl             | xhenzl04  |                                 | */
/* | Martin Herich            | xheric00  |                                 | */
/* | Maros  Barjak            | xbarja00  |                                 | */
/* \------------------------------------------------------------------------/ */
       
/*
 *		Hlavny program
 */
 
// ovladanie debugovacich vypisov
#define DEBUG_ON 0

#include "str.h"
#include "parser.h"
#include "scanner.h"
#include "hash.h"
#include "inst_lst.h"
#include "error.h"
#include "interpret.h"

#define SYMBOL_TABLE_SIZE 10000 // velkost hashovacej tabulky

static tSymbolTable *_st=NULL;
static ilist_t *_instrList=NULL;

static void __cleanup()
{
	if(_st != NULL) {
		htable_dispose(_st);
	}
	if(_instrList != NULL) {
		ilist_free(_instrList);   // zrusenie listu instrukcii
	}
	CloseFile();
}

int main(int argc, char *argv[])
{

	atexit(__cleanup);

	if (argc == 2)
	{
		if (ReadFile(argv[1])==RUN_ERROR)
		{
			fprintf(stderr,"Chyba pri otvarani suboru !\n"); 
			return RUN_ERROR; // doslo ku chybe pri otvarani suboru
		}
	}
	else 
	{
		fprintf(stderr,"Chybne parametre !\n");
		return RUN_ERROR;
	}

	// inicializacia tabulky symbolov
	if ((_st = htable_init(SYMBOL_TABLE_SIZE))==NULL)
    {
        fprintf(stderr,"Chyba alokacie tabulky symbolov !\n");
        return RUN_ERROR;
    }
  
    if ((_instrList = ilist_new()) == NULL)
    {
        fprintf(stderr,"Chyba alokacie listu instrukcii !\n");        
        return RUN_ERROR;
    }

	// SYNTAKTICKA ANALYZA
	int result = parse(_st, _instrList); // provedeme syntaktickou analyzu

    //testovacie vypisy
	htable_write(_st);
	if (result==SYNTAX_OK)
		print_debug(("SYNTAX OK"))
	else if (result==SYNTAX_ERROR)
		print_debug(("SYNTAX_ERROR"))
    else if (result==SEM_ERROR)
		print_debug(("SEM_ERROR"))
    else if (result==RUN_ERROR)
        fprintf(stderr,"Chyba pri alokacii pamate !\n");        
	if (result!=SYNTAX_OK) return result;

	// INTERPRETACIA 
    print_debug(("Zacina INTERPRETACIA"))
	if (interpret_run(_instrList) != 0) 
    {
        fprintf(stderr,"Chyba pri interpretacii !");                
        return INT_ERROR;
    }

    // vypis tabulky symbolov po interpretacii
    print_debug(("Koniec INTERPRETACIE"))
	htable_write(_st);

	return result;
}
