/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
/*      ###           Implementacia interpretu jazyka IFJ09          ###      */  
/* ########################################################################## */
/*  Akademicky rok: 2009/2010 ## Predmet: IFJ ## Datum odovzdania : 13.12 09  */
/* |------------------------------------------------------------------------| */
/* |        Team              |   Login   |       Nazov a popis modulu      | */ 
/* |------------------------------------------------------------------------| */
/* | Martin Haviar - veduci   | xhavia01  |                                 | */
/* | Jozef  Henzl             | xhenzl04  |                                 | */
/* | Martin Herich            | xheric00  |                                 | */
/* | Maros  Barjak            | xbarja00  |                                 | */
/* \------------------------------------------------------------------------/ */
                                                                 
/*******************************************************************************
 *		        Syntakticka analyza rekurzivnym zostupom                       *
 ******************************************************************************/
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "str.h"
#include "parser.h"
#include "scanner.h"
#include "hash.h"
#include "stack.h"
#include "inst_lst.h"
#define DEBUG_ON 0
#include "error.h"

#define RETURN_LEX_ERROR \
{ fprintf(stderr, "Lexikalna chyba na riadku %d\n", getCurLine()); return LEX_ERROR; }

#define RETURN_SYN_ERROR \
{ fprintf(stderr, "Syntakticka chyba na riadku %d\n", getCurLine()); return SYNTAX_ERROR; }

#define RETURN_SEM_ERROR \
{ fprintf(stderr, "Semanticka chyba na riadku %d\n", getCurLine()); return SEM_ERROR; }


tSymbolTable *table;    // tabulka symbolov
ilist_t *ilist=NULL;         // zoznam instrukcii
int token;              // aktualny token
string attr;            // atribut tokenu
int repeat_token;       // urcuje ci sa ma predosly token zopakovat

int readln_status;      // na kontrolu typov nacitavanych v readln
int repeat_counter=0;   // sluzi ako buffer zanorenia repeat-until vetiev

int counterVar = 1; // na generovanie unikatnych nazvov docasnych premennych

void generateVariable(string *var)
// generuje jedinecne nazvy identifikatoru
// nazev se sklada ze znaku $ nasledovanym cislem
// postupne se tu generuji prirozena cisla a do nazvu promenne se ukladaji
// v reverzovanem poradi - na funkcnost to nema vliv, ale je jednodussi implementace

{
  strClear(var);
  strAddChar(var, '$');
  int i;
  i = counterVar;
  while (i != 0)
  {
    strAddChar(var, (char)(i % 10 + '0'));
    i = i / 10;
  }
  counterVar ++;
} 

/*
 *  Vlozenie novej instrukcie do zoznamu instrukcii
 */
instruction_t *generateInstruction(int instType, htable_item *addr1, htable_item *addr2, htable_item *addr3)
{
    instruction_t *inst;

	assert(ilist != NULL);
	inst = ilist_addi_end(ilist);
    inst->type = instType;
    inst->i1 = addr1;
    inst->i2 = addr2;
    inst->i3 = addr3;
    return inst;
}


/*
 *  Vytvori konstantu v tabulke symbolov. Pouzite pri podmienenych skokoch
 */
htable_item *create_tmp_var(int value)
{
    htable_item *tmp;
    int result;  
    // vytvorime nazov pre docasnu premennu            
    string newVar;
    strInit(&newVar);
    generateVariable(&newVar);
    result = htable_lookup(table, &tmp, newVar.str, 1);  // 1 => ide o docasnu premennu
    tmp->type = T_INTEGER;
    tmp->data.num_int = value;
    strFree(&newVar);
    if (result==HT_ERROR) return NULL;   // chyba pri alokacii
    // pridanie instrukcie negacie
    return tmp;       
}

/*******************************************************************************
 *		Jednotlive funkcie zodpovedaju nonterminalom gramatiky
 ******************************************************************************/

/*
 *  Najvyssi nonterminal <prog>
 *  1	<prog>	 	-->	<declr> <stat_body> . <EOF>
 */
int _prog()
{
    int result;
    switch (token)
    {
    case BEGIN:     // ak deklaracia bola prazdna
    case VAR:       // normalna deklaracia
        result = _declr();
        if (result != SYNTAX_OK) return result;

        result = _stat_body();
        if (result != SYNTAX_OK) return result;

        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
        if (token != DOT_TOKEN)
            RETURN_SYN_ERROR;    // bodka za end

        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;            
        if (token != END_OF_FILE )
            RETURN_SYN_ERROR;    //  EOF 
                                  
        return SYNTAX_OK;
    }
    // ine tokeny generuju chybu
    RETURN_SYN_ERROR;
}

/*
 *  nonterminal <declr>
 *  2	<declr>	 	-->	var <declr_list>
 *  3	<declr>	 	-->	begin
 */
int _declr()
{
    if (token==BEGIN) 
    {
        // nacita dalsi token
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
        return SYNTAX_OK;
    }
    if (token==VAR) 
    {
        // nacita dalsi token
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
        return _declr_list();
    }

    RETURN_SYN_ERROR;    // ine tokeny generuju chybu
}

/*
 *  nonterminal <stat_body>
 *  7	<stat_body>	-->	end         // prazdny begin end.
 *  8	<stat_body>	-->	<stat_list>
 */
int _stat_body()
{
    // token nacitany v  <declr>
    switch (token)
    {
    case END:                       // prazdny begin end.
        return SYNTAX_OK; 
        break;
    case BEGIN: case IF: case READLN: case WHILE: case WRITE: case ID: case REPEAT:
        return _stat_list(); 
        break;
    }    
    RETURN_SYN_ERROR;   // ine tokeny generuju chybu
}

/*
 *  nonterminaly <declr_list> a <is_declr_end>
 *  4	<declr_list>	-->	"id : <type>; <is_declr_end>"
 *  5	<is_declr_end>	-->	<declr_list>
 *  6	<is_declr_end>	-->	begin
 */
int _declr_list()
{
    // token nacitany v  <declr>
    int result;
    htable_item *p;

    switch (token)
    {
    case ID:
        // pokusime sa ulozit ID do tabulky symbolov,

        result = htable_lookup(table, &p, attr.str, 0);  // 0 => nejde o docasne id

        if (result == HT_FOUND) RETURN_SEM_ERROR;        // identifikator uz existoval
        if (result == HT_ERROR) return RUN_ERROR;           // chyba pri alokacii 

        // bol pridany novy identifikator, nacitaj dvojbodku
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;        
        if (token != COLON ) RETURN_SYN_ERROR;

        // zistenie a nastavenie typu
        int id_type;
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
        if (_type(&id_type) == SYNTAX_ERROR)
            RETURN_SYN_ERROR;
        p->type = id_type;
        
        if (p->type == T_STRING) strInit(&p->data._string);

        // nasledovat by mala ";"
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
        if (token != SEMICOLON ) RETURN_SYN_ERROR;

        // podla dalsieho tokenu zistime ci deklaracie pokracuju
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;

        if (token==BEGIN) 
        {
            // nacita dalsi token za beginom
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            return SYNTAX_OK;               // koniec deklaracii
        }
        if (token==ID) return _declr_list();    // deklaracie pokracuju
        break;
    }    
    RETURN_SYN_ERROR;
}

/*  parametre funkcie readln(<id_list>)
 *  16	<id_list>	-->	id <end_id_list>
 *  17	<end_id_list>	-->	, <id_list>
 *  18	<end_id_list>	-->	)       // prava zatvorka
 */
int _id_list()
{
    if (token != ID ) RETURN_SYN_ERROR;
    // zisti odkaz v tabulke symbolov na toto id
    htable_item *p;
    int result = htable_lookup(table, &p, attr.str, 0);
    if (result!=HT_FOUND) RETURN_SEM_ERROR;     // ak sa tam dane id nenachadzalo 

    // kontrola typov - readln(string) || readln([int,double]*)
    // readln_status : 0 prve id, 1 predchadzal string, 2 predchadzalo int alebo double 
    if (p->type==T_STRING)
    {
        if (readln_status!=0) RETURN_SYN_ERROR;  // string moze byt len na zaciatku
        readln_status=1;    // posledny read bol string
    }
    else
    {
        // prislo cislo int || double - pokracuj len ak nepredchadzal string
        if (readln_status==1) RETURN_SYN_ERROR;  // ak predchadzal string => chyba
        readln_status = 2;  // posledny read bolo cislo
    }

    // vygenerujeme instrukciu s adresou do ktorej sa ma nacitat
    generateInstruction(INST_READ, p, NULL, NULL);

    //podla nasledujuceho tokenu pokracujeme v nacitavani id alebo skoncime
    if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
    if (token==COMMA)   // ciarka -> pokracujeme
    {
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
        return _id_list();
    }
    if (token==RIGHT_BRACKET) {
		generateInstruction(INST_CLR, NULL, NULL, NULL);
		return SYNTAX_OK;
	}
    RETURN_SYN_ERROR;    // ine tokeny generuju chybu
}

/**************************************************************************************************/
// globalne zasobniky pouzite v expr
    stack_t zasobnik;             // zasobnik na symboly
   // stack_t brack_stack;

// funkcia sa vola atexit() aby uvolnila zasobniky v expr
void dispose_stacks()
{
    print_debug((">>> Uvolnujem zasobniky <<<"))
    stackDispose(&zasobnik);
    //stackDispose(&brack_stack);
}

/*
 *  spracovanie vyrazov analyzou zdola nahor
 *  cez result sa vrati odkaz na vyslednu hodnotu vyrazu
 */
int _expr(htable_item **result)
{    
    print_debug(("\nSpustam PA"))   

    string newVar;                // generovany nazov cislenej/retazcovej konstanty do ht

    int vstup, in_stack;          // konvertovany token zo vstupu, symbol v zasobniku
    int is_nonterm = FALSE;       // prepinac urcujuci ci je bol najdeny E symbol v zasobniku,
                                  //   (kvoli rozliseniu vyhodnoteneho vyrazu ($E$) a prazdneho
                                  //    vyrazu ($$)
    int tmp_symbol1,              // pomocne premenne do kt. sa nacitavaju symboly zo zasobniku pri 
        tmp_symbol2,              //   redukcii vyrazu podla pravidiel 
        tmp_symbol3,
        ht_state,
        //is_bracket,
        lbrc_counter = 0;         // pocitadlo lavych zatvoriek
    
    htable_item *ht_item1,        // pomocne premenne s pointermi na polozky v hashke
                *ht_item2,
                *ht_item3,
                *ht_item_result;
                
    char precedencia;             // symbol v PT
    char* ptable[14] =            // Precedencna tabulka
    {//   div  <> ==
     //+-* ()i= <><>$
      ">><<<><>>>>>>>", //+
      ">><<<><>>>>>>>", //-
      ">>>><><>>>>>>>", //*
      ">>>><><>>>>>>>", //div
      "<<<<<=<<<<<<<0", //(
      ">>>>0>0>>>>>>>", //)
      ">>>>0>0>>>>>>>", //i
      "<<<<<><>><<<<>", //=
      "<<<<<><>><<<<>", //<>
      "<<<<<><>>>>>>>", //<
      "<<<<<><>>>>>>>", //>
      "<<<<<><>>>>>>>", //<=
      "<<<<<><>>>>>>>", //>=
      "<<<<<0<<<<<<<0", //$
    };
    
    // inicializujeme zasobnik, a pushneme tam $
    stackInit(&zasobnik);
    stackPush(&zasobnik, PA_DOLLAR, NULL);
/*    
    stackInit(&brack_stack);
    stackPush(&brack_stack, PA_DOLLAR, NULL);
*/    
    print_debug(("Vytvaram zasobniky"))
    // hlavny algoritmus PA
    // token mame nacitany uz pre volanim funkcie ...
    do
    {
      ht_item1 = ht_item2 = ht_item3 = ht_item_result = NULL;
      /*
      *   konverzia tokenov na symboly v precedencnej tabulke
      */
      print_debug(("-------------------"))
      switch (token)
      {  
        // binarne operatory: + - / div
        case PLUS_TOKEN:    vstup = PA_PLUS;      break;
        case MINUS_TOKEN:   vstup = PA_MINUS;     break;
        case MULT_TOKEN:    vstup = PA_MULT;      break;
        case DIV:           vstup = PA_DIV;       break;
        
        // zatvorky: ( )
        case LEFT_BRACKET:  vstup = PA_LBRACKET;  break;  // tu sa moze zvysovat, lebo ( sa vzdy shifuje
        
        case RIGHT_BRACKET: vstup = PA_RBRACKET;  break;
        
        // identifikatory: a 6 6.6 6e10 "bla"
        case ID:            vstup = PA_ID;        break;
        case INT_CONST:     vstup = PA_ID;        break;
        case DOUBLE_DOT:    vstup = PA_ID;        break;
        case DOUBLE_EXP:    vstup = PA_ID;        break;
        case STRING:        vstup = PA_ID;        break;
        
        // porovnavanie: = < <= <> > >=
        case EKVIVALENCE:   vstup = PA_EQ;        break;
        case LESS_TOKEN:    vstup = PA_LESS;      break;
        case LESS_EQUAL:    vstup = PA_LESSEQ;    break;
        case UN_EQUAL:      vstup = PA_NEQ;       break;
        case GREATER_TOKEN: vstup = PA_GRE;       break;
        case GREATER_EQUAL: vstup = PA_GREEQ;     break;
        
        // ukoncovanie tokeny: ; do else then end
        case SEMICOLON:     vstup = PA_DOLLAR;    break;
        case DO:            vstup = PA_DOLLAR;    break;
        case ELSE:          vstup = PA_DOLLAR;    break;
        case THEN:          vstup = PA_DOLLAR;    break;
        case END:           vstup = PA_DOLLAR;    break;
        case COMMA:         vstup = PA_DOLLAR;    break;        
        case UNTIL:         vstup = PA_DOLLAR;    break;        
        
        // prislo nieco co nemalo -> SYN ERR
        default:  
          print_debug(("prislo nieco co nemalo -> SYN ERR"))
          dispose_stacks();
          RETURN_SYN_ERROR;
      }  
      
      // kontrola zatvoriek pomocou stacku - aby sa nespracovala zatvorka ktora
      // uz nepatri do vyrazu
//      if (vstup == PA_LBRACKET)
//        stackPush(&brack_stack, PA_LBRACKET, NULL);
      /*  
      print_debug(("lbrc_counter == %i\n", lbrc_counter))  
      if (vstup == PA_RBRACKET)
      {
        //is_bracket = stackTopPop(&brack_stack, &ht_item1);
        
        print_debug(("..is bracket == %i\n", is_bracket))
        if (is_bracket == PA_DOLLAR)
        {
          vstup = PA_DOLLAR;
          stackPush(&brack_stack, PA_DOLLAR, NULL); // vzdy je treba aby ostal aspon jeden $
        }      
      }*/
      
      // najdeme si navyssi terminalny symbol v zasobniku
      in_stack = stackGetActSucc(&zasobnik);
      if (in_stack == PA_NONTERMINAL)
      {
        print_debug(("..NONTERM true"))
        is_nonterm = TRUE;
        in_stack = stackGetActSucc(&zasobnik);
      }  
      print_debug(("Na zasobniku je %i", in_stack))
      print_debug(("Na vstupe je %i", vstup))

      // ak je na stacku $E a na vstupe $, mame vyhodnoteny vyraz a vyskocime z cyklu
      if (((in_stack == PA_DOLLAR) && (is_nonterm == TRUE)) && (vstup == PA_DOLLAR))
      {
        print_debug(("KONIEC spracovania vyrazu"))
        stackTopPop(&zasobnik, &ht_item_result);
        (*result) = ht_item_result;
        print_debug(("result type: %i\n\n", ht_item_result->type))
        break;
      }
      if (((in_stack == PA_DOLLAR) && (is_nonterm == TRUE)) && (vstup == PA_RBRACKET))
      {
        print_debug(("........vstup: %i",vstup))
        print_debug(("........counter: %i",lbrc_counter))
        print_debug(("KONIEC spracovania vyrazu s nabytocnou )"))
        stackTopPop(&zasobnik, &ht_item_result);
        (*result) = ht_item_result;
        print_debug(("result type: %i\n\n", ht_item_result->type))
        break;
      }
      // zistime si precedenciu
      precedencia = ptable[in_stack][vstup];
      //////////////////////////////////////////////////////////////////////////////////////////////
      if ((precedencia == '<') || (precedencia == '='))
      // shifujeme na zasobnik
      // a spracujeme identifikator
      {
          char *end_ptr;  // pomocna premenna pre strtol()/strtod()
          int i_val;      // pomocna premenna pre strtol()
          double d_val;   // pomocna premenna pre strtod()
                              
          print_debug(("..Shift na zasobnik"))
          if (token == ID)  // premenna
          {
            print_debug(("....Spracuvam identifikator '%s'", attr.str))
            // vyhladame v hashke premennu
            ht_state = htable_lookup(table, &ht_item1, attr.str, 0);
            // ak sa nevydarila alokacia
            if (ht_state == HT_ERROR)
            {
              dispose_stacks();
              return RUN_ERROR;
            }
            // ak sa premenna nenasla (je nedeklarovana)
            else if (ht_state == HT_NOT_FOUND)
            {
              print_debug(("NEDEKLAROVANA premenna"))
              dispose_stacks();
              RETURN_SEM_ERROR;
            }
            print_debug(("....Identifikator OK"))
            // ak najde je vsecko okej
          }
          //////////////////////////////////////////////////////////////////////////////////////////
          else if (token == INT_CONST)  // cislo typu int
          {
            print_debug(("....Spracuvam INT '%s'", attr.str))
            // vytvorime novy flek v ht
            strInit(&newVar);
            generateVariable(&newVar);
            ht_state = htable_lookup(table, &ht_item1, newVar.str, 0);
            strFree(&newVar);
            // ak sa nevydarila alokacia alebo jedinecny nazov existuje (WTF?)
            if ((ht_state == HT_ERROR) || (ht_state == HT_FOUND))
            {
              dispose_stacks();
              return RUN_ERROR;
            }
            // ak sa konstanta vytvorila, ulozime jej udaje do ht 
            else if (ht_state == HT_INSERTED)
            {
              ht_item1->type = T_INTEGER;
               errno = 0;
               i_val = strtol(attr.str, &end_ptr, 0);
               /* Check for various possible errors */

               if (errno == ERANGE)
               {
                    perror("strtol");
                    dispose_stacks();
                    RETURN_LEX_ERROR;
               }

               /* If we got here, strtol() successfully parsed a number */
              ht_item1->data.num_int = i_val;
              print_debug(("....INT spracovany"))
            }
          }
          //////////////////////////////////////////////////////////////////////////////////////////
          else if ((token == DOUBLE_DOT) || (token == DOUBLE_EXP))  // cislo typu double
          {
            print_debug(("....Spracuvam DOUBLE '%s'", attr.str))
            // vytvorime novy flek v ht
            strInit(&newVar);
            generateVariable(&newVar);
            ht_state = htable_lookup(table, &ht_item1, newVar.str, 0);
            strFree(&newVar);
            // ak sa nevydarila alokacia alebo jedinecny nazov existuje (WTF?)
            if ((ht_state == HT_ERROR) || (ht_state == HT_FOUND))
            {
              dispose_stacks();
              return RUN_ERROR;
            }
            // ak sa konstanta vytvorila, ulozime jej udaje do ht 
            else if (ht_state == 0)
            {
              ht_item1->type = T_DOUBLE;
              
                d_val = strtod(attr.str, &end_ptr);
               /* Check for various possible errors */

               if (errno == ERANGE)
               {
                    perror("strtod");
					stackDispose(&zasobnik);
                    RETURN_LEX_ERROR;
               }

              ht_item1->data.num_double = d_val;
              print_debug(("....DOUBLE spracovany"))
            }
          }
          //////////////////////////////////////////////////////////////////////////////////////////
          else if (token == STRING) // retazec
          {
            print_debug(("....Spracuvam String '%s'", attr.str))
            strInit(&newVar);
            generateVariable(&newVar);
            ht_state = htable_lookup(table, &ht_item1, newVar.str, 0);
            strFree(&newVar);
            // ak sa nevydarila alokacia alebo jedinecny nazov existuje (WTF?)
            if ((ht_state == HT_ERROR) || (ht_state == HT_FOUND))
            {
              dispose_stacks();
              return RUN_ERROR;
            }
            // ak sa konstanta vytvorila, ulozime jej udaje do ht 
            else if (ht_state == HT_INSERTED)
            {
              // inicializacia pomocneho stringu a kopirovanie obsahu z attr
              ht_item1->type = T_STRING;
              strInit(&ht_item1->data._string);              
              strConcat(&ht_item1->data._string, attr.str);

              print_debug(("....String je '%s'", ht_item1->data._string.str))
              print_debug(("....String spracovany"))
            }
          }
        stackPush(&zasobnik, vstup, ht_item1);
        
        if (token==RIGHT_BRACKET)
          lbrc_counter++;
        if (token==RIGHT_BRACKET)
          lbrc_counter--;
        
        token = getToken(&attr);
   
        if (token==WRONG_LEXEM) RETURN_LEX_ERROR;
        
        print_debug(("..Koniec shiftovania"))
      }
      //////////////////////////////////////////////////////////////////////////////////////////////
      else if (precedencia == '>')
      // redukujeme vyraz
      // E -> i
      // E -> (E)
      // E -> E + E
      // E -> E - E
      // E -> E * E
      // E -> E div E
      // E -> E = E
      // E -> E < E
      // E -> E > E
      // E -> E <> E
      // E -> E <= E
      // E -> E >= E
      {
        print_debug(("..Zacinam redukciu vyrazu"))
        // nacitame si pravu stranu pravidla
        tmp_symbol1 = stackTopPop(&zasobnik, &ht_item1);
        
        // a podla zaciatku urcime pravidlo [i, E, )]
        // INDENTIFIKATOR
        if (tmp_symbol1 == PA_ID)
        {
        // E -> i //////////////////////////////////////////////////////////////////////////////////
          print_debug(("....E -> i"))
          stackPush(&zasobnik, PA_NONTERMINAL, ht_item1);
          print_debug(("......vstup == %i",vstup))
        }
        // ZATVORKA [ E -> (E) ]
        else if (tmp_symbol1 == PA_RBRACKET) 
        {
        // E -> (E) ////////////////////////////////////////////////////////////////////////////////  
          print_debug(("....E -> (E)"))
          // popneme si zo zasobniku dalsi symbol (musi to byt NONTERM)
          tmp_symbol2 = stackTopPop(&zasobnik, &ht_item2);
          if (tmp_symbol2 != PA_NONTERMINAL)
          {
            print_debug(("SYN chyba v E -> (E): non E"))            
            dispose_stacks();
            RETURN_SYN_ERROR;
          }
          // popneme si zo zasobniku dalsi symbol (musi to byt lava zatvorka)
          tmp_symbol3 = stackTopPop(&zasobnik, &ht_item3);
          if (tmp_symbol3 != PA_LBRACKET)
          {
            print_debug(("tmpsym3 je: %i",tmp_symbol3))
            print_debug(("SYN chyba v E -> (E): non ("))
            dispose_stacks();
            RETURN_SYN_ERROR;
          }
          
          // a zredukujeme na E 
          stackPush(&zasobnik, PA_NONTERMINAL, ht_item2);
          print_debug(("......Redukcia OK"))
          /*
          print_debug(("........vstup: %i\n",vstup))
          print_debug(("........counter: %i\n",lbrc_counter))
          if ((vstup == PA_RBRACKET) && (lbrc_counter == 0))
          {
            print_debug(("KONIEC spracovania vyrazu s nabytocnou )\n"))
            stackTopPop(&zasobnik, &ht_item_result);
            (*result) = ht_item_result;
            print_debug(("result type: %i\n\n", ht_item_result->type))
            break;
          }*/
        }
        ////////////////////////////////////////////////////////////////////////////////////////////
        else if (tmp_symbol1 == PA_NONTERMINAL) // prava stana pravidla je neterminalny symbol E
        {
          print_debug(("....E -> E x x"))
          // popneme si zo stacku symbol a podla neho urcime co sa bude robit
          tmp_symbol2 = stackTopPop(&zasobnik, &ht_item2);
          // a hned si popneme aj treti symbol, co moze byt len NONTERMINAL E
          tmp_symbol3 = stackTopPop(&zasobnik, &ht_item3);
          if (tmp_symbol3 != PA_NONTERMINAL)
          {
            print_debug(("SYNERR treti popnuty znak neni E"))
            dispose_stacks();
            RETURN_SYN_ERROR;
          }
          /* ***************************************************************************************
          *  SEMANTIKA binarnich operacii           
          *************************************************************************************** */
          strInit(&newVar);
          generateVariable(&newVar);
          ht_state = htable_lookup(table, &ht_item_result,newVar.str, 1);
          strFree(&newVar);
          
            if ((ht_state == HT_ERROR) || (ht_state == HT_FOUND))
            {
              dispose_stacks();
              return RUN_ERROR;
            }
            // ak sa konstanta vytvorila je to ok 

          // + - *
          if ((tmp_symbol2 >= 0) && (tmp_symbol2 <= 2))
          {
            // INT (+|-|*) INT = INT
            if ((ht_item3->type == T_INTEGER) && (ht_item1->type == T_INTEGER))
              ht_item_result->type = T_INTEGER;
            // DOUBLE (+|-|*) INT = DOUBLE
            else if ((ht_item3->type == T_DOUBLE) && (ht_item1->type == T_INTEGER))
              ht_item_result->type = T_DOUBLE;
            // INT (+|-|*) DOUBLE = DOUBLE
            else if ((ht_item3->type == T_INTEGER) && (ht_item1->type == T_DOUBLE))
              ht_item_result->type = T_DOUBLE;
            // DOUBLE (+|-|*) DOUBLE = DOUBLE
            else if ((ht_item3->type == T_DOUBLE) && (ht_item1->type == T_DOUBLE))
              ht_item_result->type = T_DOUBLE;
            // STRING + STRING = STRING (konkatenacia)
            else if ((tmp_symbol2 == PA_PLUS) && (ht_item3->type == T_STRING) && (ht_item1->type == T_STRING))
            {
              ht_item_result->type = T_STRING;
              strInit(&ht_item_result->data._string);
            }
            // nepovolene kombinacie vratia semanticku chybu
            else
            {
              print_debug(("Nepovolena kombinacia typov pri bin operaciach MAT"))
              dispose_stacks();
              RETURN_SEM_ERROR;
            }     
          }
          // div
          else if (tmp_symbol2 == 3)
          {
            // INT div INT = INT
            if ((ht_item3->type == T_INTEGER) && (ht_item1->type == T_INTEGER))
              ht_item_result->type = T_INTEGER;
            else
                {
                    print_debug(("Nepovolene typy pri div"))
                    dispose_stacks();
                    RETURN_SEM_ERROR;
                }
          }
          // porovnania 
          else if ((tmp_symbol2 >= 7) && (tmp_symbol2 <= 12))
          {
            // INT (=|<>|<|>|<=|>=) INT = INT
            if ((ht_item3->type == T_INTEGER) && (ht_item1->type == T_INTEGER))
              ht_item_result->type = T_INTEGER;
            // DOUBLE (=|<>|<|>|<=|>=) DOUBLE = INT
            else if ((ht_item3->type == T_DOUBLE) && (ht_item1->type == T_DOUBLE))
              ht_item_result->type = T_INTEGER; 
            // STRING (=|<>|<|>|<=|>=) STRING = INT
            else if ((ht_item3->type == T_STRING) && (ht_item1->type == T_STRING))
              ht_item_result->type = T_INTEGER;
            else
            {
              print_debug(("Nepovolena kombinacia typov pri bin operaciach CMP"))
              dispose_stacks();
              RETURN_SEM_ERROR;
            }   
          }
          else
          {
            // ak mame korektne operatory -> SEM ERR
            // nekorektne op. sa vyhodnotia v nasledujucom vetveni
            if (((tmp_symbol2 >= 0) && (tmp_symbol2 <= 4)) || ((tmp_symbol2 >= 7) && (tmp_symbol2 <= 12)))  
            {  
              print_debug(("Nepovolena kombinacia typov pri bin operaciach"))
              dispose_stacks();
              RETURN_SEM_ERROR;
            }
          }
          /****************************************************************************************/
        
        // E -> E + E
          if (tmp_symbol2 == PA_PLUS)
          {
            
            generateInstruction(INST_ADD, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E + E"))
          }
        // E -> E - E
          else if (tmp_symbol2 == PA_MINUS)
          {
            
            generateInstruction(INST_SUB, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E - E"))          
          }
        // E -> E * E
          else if (tmp_symbol2 == PA_MULT)
          {
            
            generateInstruction(INST_MUL, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E * E"))          
          }
        // E -> E div E
          else if (tmp_symbol2 == PA_DIV)
          {
             
            generateInstruction(INST_DIV, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E div E"))         
          }
        // E -> E = E
          else if (tmp_symbol2 == PA_EQ)
          {
            
            generateInstruction(INST_CMP_EQUAL, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E = E"))          
          }
        // E -> E <> E
          else if (tmp_symbol2 == PA_NEQ)
          {
            
            generateInstruction(INST_CMP_NOTEQUAL, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E <> E"))          
          }
        // E -> E < E
          else if (tmp_symbol2 == PA_LESS)
          {
            
            generateInstruction(INST_CMP_LESS, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E < E"))          
          }  
        // E -> E <= E
          else if (tmp_symbol2 == PA_LESSEQ)
          {
            
            generateInstruction(INST_CMP_LESSEQ, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E <= E"))          
          }
        // E -> E > E
          else if (tmp_symbol2 == PA_GRE)
          {
            
            generateInstruction(INST_CMP_GREAT, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E > E"))          
          }
        // E -> E >= E 
          else if (tmp_symbol2 == PA_GREEQ)
          {
            
            generateInstruction(INST_CMP_GREATEQ, ht_item3, ht_item1, ht_item_result);
            stackPush(&zasobnik, PA_NONTERMINAL, ht_item_result);
            print_debug(("....E -> E >= E"))          
          }
        // prislo nieco co nemalo
          else
          {
            print_debug(("SYNERR druhy znak neni operator"))
            dispose_stacks();
            RETURN_SYN_ERROR;          
          }
        ////////////////////////////////////////////////////////////////////////////////////////////          
        }
        else  // ani neterminalny symbol, ani identifikator, ani zatvorka = synt. chyba
        {
          print_debug(("SYNERR prvy znak neni E,i,)"))
          dispose_stacks();
          RETURN_SYN_ERROR;
        }
      }
      else if (precedencia == '0')
      // syntakticka chyba
      {
        print_debug(("SYNERR precedencia 0"))
        dispose_stacks();
        RETURN_SYN_ERROR;
      }
      else
      // neznama chyba (neznamy znak v PT)
      {
        print_debug(("RUNERR Zly znak v PT"))
        dispose_stacks();
        return RUN_ERROR;
      }
    }
    while (1);
    dispose_stacks();
//    dispose_stacks();
//    stackDispose(&brack_stack);
    print_debug(("aktualny token je %i", token))
    RepeatLastToken();
    
    return SYNTAX_OK;
}
/**************************************************************************************************/
/*  parametre funkcie write(<id_list>)
 *  <expr_list>	-->	<expr> <end_expr_list>
 *  21	<end_expr_list>	-->	, <expr_list>
 *  22	<end_expr_list>	-->	)       // prava zatvorka 
 */
int _expr_list()
{
    int result;
    htable_item *expr_result;

    result = _expr(&expr_result);
    //generovanie instrukcie pre write
    generateInstruction(INST_WRITE, expr_result, NULL, NULL);

    if (result!=SYNTAX_OK) return result;

    // podla dalsieho tokenu sa rozhodneme ci pokracovat
    if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
    if (token==RIGHT_BRACKET) return SYNTAX_OK;
    if (token==COMMA)
    {
        if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
        return _expr_list();
    }    
    RETURN_SYN_ERROR;    // ine tokeny generuju chybu    
}

/*
 *  nonterminal <stat>
 *  15	<stat>	 	-->	readln ( <id_list>
 *  19	<stat>	 	-->	write ( <expr_list>
 *  23	<stat>	 	-->	begin <is_empty>
 *  24	<is_empty>	-->	end          // prazdna podvetva
 *  25	<is_empty>	-->	<stat_list>
 *  26	<stat>	 	-->	if <expr> then <stat> <if_else>
 *  27	<if_else>	-->	else <stat>
 *  28	<if_else>	-->	epsilon
 *  29	<stat>	 	-->	while <expr> do <stat>
 *  30	<stat>	 	-->	id := <inner>
 *  34	<stat>	 	-->	repeat <repeat_body> <expr>
                        - repeat_body je implementovane pomocou stat_listu
 */
int _stat()
{
    int result;
    htable_item *expr_result;
    htable_item *id;
    htable_item *a, *b;
    instruction_t * addrOfIfGoto; 

    switch (token)
    {
        case READLN:
           // za readln sa ocakava "("
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;           
            if (token != LEFT_BRACKET) 
                RETURN_SYN_ERROR;            
            // ocakava zoznam id
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            readln_status=0;    // nasleduje prve id
            return _id_list();

        case WRITE:
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            if (token  != LEFT_BRACKET)
                RETURN_SYN_ERROR;
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            return _expr_list();

        case BEGIN:
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            if (token == END ) 
                return SYNTAX_OK;
            else
                return _stat_list();    // neprazdna vetva programu

        case IF:
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            // vyhodnotenie vyrazu
            result = _expr(&expr_result);
            if (result!=SYNTAX_OK) return result;
            // semanticka kontrola vyrazu (musi byt typu integer)
            if (expr_result->type != T_INTEGER) RETURN_SEM_ERROR;
            

            // nagenerujeme instrukciu podmieneneho skoku a ulozime jej adresu,
            // pretoze bude treba doplnit adresu skoku            
            assert(expr_result);

            addrOfIfGoto = generateInstruction(INST_JZ, expr_result, NULL, NULL); 

            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            if (token!=THEN) RETURN_SYN_ERROR;
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            result = _stat();
            
            // nagenerujeme skok na end_if, neskor donho nastavime adresu
            instruction_t *goto_end;
            goto_end = generateInstruction(INST_JMP, NULL, NULL, NULL);            

            // nagenerujeme label else_if, kam sa skoci ak expr==0
            // a ulozime si cislo instrukcie
            generateInstruction(INST_LABEL, NULL, NULL, NULL);
            il_addr_t else_if = ilist_last_addr(ilist);

            // podla dalsieho tokenu zistime ci nasleduje aj else vetva
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            if (token==ELSE)
            {
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                result = _stat();
                if (result!=SYNTAX_OK) return result;
            }
            else RepeatLastToken(); // bez else vetvy -> zopakuje posledny token

            // nagenerujeme label end_if, kam sa skoci po prevedeni prikazov pre expr!=0
            // a ulozime si cislo instrukcie
            generateInstruction(INST_LABEL, NULL, NULL, NULL);
            il_addr_t end_if = ilist_last_addr(ilist);

            // nastavenie adries predchadzajucich skokov
            // je potrebne vytvorit polozky v symbol table aby sme vedeli kam skakat
            goto_end->i3 = create_tmp_var(end_if);
            addrOfIfGoto->i3 = create_tmp_var(else_if);
            if (goto_end->i3==NULL || addrOfIfGoto->i3==NULL) 
                return RUN_ERROR;   // chyba pri alokacii pomocnych konstant
            return SYNTAX_OK;

        case WHILE:     // while <expr> do <stat>
            // nagenerujeme instrukciu label ako zaciatok while a zapamatame si ju
            generateInstruction(INST_LABEL, NULL, NULL, NULL);
            il_addr_t while_start = ilist_last_addr(ilist);

            // nacitame dalsi token a vyhodnotime vyraz, ktory vrati vysledok 
            // ako odkaz do tabulky symbolov
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;            
            result = _expr(&expr_result);
            if (result!=SYNTAX_OK) return result;
            // semanticka kontrola vyrazu (musi byt typu integer)
            if (expr_result->type != T_INTEGER) RETURN_SEM_ERROR;             

            // nagenerujeme instrukciu podmieneneho skoku a ulozime jej adresu,
            // pretoze bude treba doplnit adresu skoku
            assert(expr_result);
            addrOfIfGoto = generateInstruction(INST_JZ, expr_result, NULL, NULL); 

            /* spracujeme "do" a telo prikazu while
             * ratame s tym ze _expr nenacital dalsi token !!! 
             *  => na konci _expr() treba pouzit RepeatLastToken()
             */
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            if (token!=DO) RETURN_SYN_ERROR;
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            result = _stat();
            if (result!=SYNTAX_OK) return result;

            // nagenerujeme instrukciu skoku
            generateInstruction(INST_JMP, NULL, NULL, create_tmp_var(while_start) );

            // nagenerujeme instrukciu druheho navestia a ulozime si cislo instrukcie
            generateInstruction(INST_LABEL, NULL, NULL, NULL);
            il_addr_t while_end = ilist_last_addr(ilist);

            // uz vieme adresu druheho navestia, nastavime teda koncovu adresu
            // v podmienenom skoku addrOfIfGoto kam sa skoci ak nebude platit expr
            // Je potrebne vytvorit polozku v tab.symbolov s cislom instrukcie
            addrOfIfGoto->i3 = create_tmp_var(while_end);

            return SYNTAX_OK;

        case ID:     //  <stat>	-->	id := <inner>
            // ulozime si odkaz na cielove id
            result = htable_lookup(table, &id, attr.str, 0);
            if (result!=HT_FOUND) RETURN_SEM_ERROR; 
            // nacitame priradenie :=
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            if (token!=ASSIGNING) RETURN_SYN_ERROR;
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;

            // typova kontrola
            if (token==SORT && id->type != T_STRING) RETURN_SEM_ERROR;
            if (token==FIND && id->type != T_INTEGER) RETURN_SEM_ERROR; 

            if (token==SORT)    // sort ( expr )
            {
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                if (token!=LEFT_BRACKET) RETURN_SYN_ERROR;
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                
                result = _expr(&expr_result);
                if (result!=SYNTAX_OK) return result;

                //generovanie instrukcie         
                if (expr_result->type != T_STRING) RETURN_SEM_ERROR; // musi byt string
                // v expr_result je odkaz na string ktory chceme zoradit, v id bude vysledok
                generateInstruction(INST_SORT, expr_result, NULL, id); 
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                if (token!=RIGHT_BRACKET) RETURN_SYN_ERROR;
                return SYNTAX_OK;

            } else
            if (token==FIND)    //  find ( expr , expr )
            {
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                if (token!=LEFT_BRACKET) RETURN_SYN_ERROR;

                // nacitanie prveho parametru
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                result = _expr(&a);
                if (result!=SYNTAX_OK) return result;
                if (a->type != T_STRING) RETURN_SEM_ERROR; // musi byt string 
                
                // nacitanie ciarky
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                if (token!=COMMA) RETURN_SYN_ERROR;

                // nacitanie druheho parametru
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                result = _expr(&b);
                if (result!=SYNTAX_OK) return result;
                if (b->type != T_STRING) RETURN_SEM_ERROR; // musi byt string
                
                //generovanie instrukcie         
                generateInstruction(INST_FIND, a, b, id);  // vysledok sa ulozi do id
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                if (token!=RIGHT_BRACKET) RETURN_SYN_ERROR;
                return SYNTAX_OK;

            } else
            {   // <expr>
                // prvy token expr je uz nacitany
                result = _expr(&expr_result);
                if (result!=SYNTAX_OK) return result;
                // typova kontrola
                if (id->type != expr_result->type) RETURN_SEM_ERROR;
                // generovanie instrukcie, vysledok sa ulozi do id
                generateInstruction(INST_MOV, expr_result, NULL, id);
                return SYNTAX_OK;
            }
            break;
            
        case REPEAT:
            // nagenerujeme instrukciu label ako zaciatok repeat a zapamatame si ju
            generateInstruction(INST_LABEL, NULL, NULL, NULL);
            il_addr_t repeat_start = ilist_last_addr(ilist);

            // nacitanie tela repeat-until pomocou _stat_list, treba zvysit repeat_counter
            // a po vykonani stat_listu treba skontrolovat ci skoncil until-om
            int _rc = ++repeat_counter;     
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            result = _stat_list();
            if (result!=SYNTAX_OK) return result;
            if (_rc-1!=repeat_counter) RETURN_SYN_ERROR; // kontrola ci prisiel until
            
            // nacitame dalsi token a vyhodnotime vyraz, ktory vrati vysledok 
            // ako odkaz do tabulky symbolov
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;            
            result = _expr(&expr_result);
            if (result!=SYNTAX_OK) return result;
            // semanticka kontrola vyrazu (musi byt typu integer)
            if (expr_result->type != T_INTEGER) RETURN_SEM_ERROR;             

            // nagenerujeme instrukciu podmieneneho skoku, skakat sa bude az kym sa nesplni podmienka
            assert(expr_result);
            addrOfIfGoto = generateInstruction(INST_JZ, expr_result, NULL, create_tmp_var(repeat_start)); 
                
            return SYNTAX_OK;
            

    }   // switch (token)
    RETURN_SYN_ERROR;
}

/*
 *  nonterminaly <stat_list> a <is_end>
 *  12	<stat_list>	-->	<stat> <is_end>
 *  13	<is_end>	-->	"; <stat_list>"
 *  14	<is_end>	-->	end
 */
int _stat_list()
{
    int result;
    switch (token)
    {
        case BEGIN: 
        case IF: 
        case READLN: 
        case WHILE: 
        case WRITE: 
        case ID:
        case END:
        case REPEAT:
           if ((result = _stat()) != SYNTAX_OK) return result;
            // dalej sa pokracuje podla dalsieho tokenu
            if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
            if (token==SEMICOLON) 
            {
                //nacita dalsi token a pokracuje v _stat_list
                if ((token = getToken(&attr)) == WRONG_LEXEM) RETURN_LEX_ERROR;
                return _stat_list();
            }
            if (token==END) return SYNTAX_OK;
            
            // taktiez moze ist o koniec repeat-until vetvy
            if (token==UNTIL) 
            {
                if (repeat_counter>0)   // untilu predchadzal repeat
                {
                    repeat_counter--;
                    return SYNTAX_OK;
                }
                else RETURN_SYN_ERROR;  // untilu nepredchadzal repeat
            }
    }    
    RETURN_SYN_ERROR;
}


int _type(int *type)
{
    switch (token)
    {
        case INTEGER:
            *type = T_INTEGER;
            return SYNTAX_OK;
        case DOUBLE:
            *type = T_DOUBLE;
            return SYNTAX_OK;
        case STRING:
            *type = T_STRING;
            return SYNTAX_OK;
    }
    RETURN_SYN_ERROR;
}


/*
 *  Funkcia vrati dalsi token zo vstupneho suboru, pricom ak je repeat_token!=0,
 *  tak nenacita novy token ale vrati aktualny a nastavi repeat_token na 0.
 *  Zavedene koli epsilon pravidlu a expr, ktore niekedy potrebuju vratenie tokenu
 */
int getToken(string_t *_attr)
{
    if (repeat_token!=0)
    {   
        repeat_token=0;
        return token;   // vrati aktualny token
    }
    else
    {   
        // nacitaj dalsi token
        return getNextToken(_attr);
    }
}

/*
 *  Pri dalsom volani getToken() sa posledny token zopakuje
 */
void RepeatLastToken()
{
    repeat_token = 1;
}

/*
 * Hlavna funkcia syntaktickej analyzy, ktora si vola nonterminaly ako funkcie
 */
int parse(tSymbolTable *ST, ilist_t *instrList)
{
//    atexit(dispose_stacks); // uvolnenie zasobnikov pri ukonceni programu
    int result;
    table = ST;
    ilist = instrList;
    strInit(&attr);
    repeat_token = 0;
    if ((token = getToken(&attr)) == LEX_ERROR)
        // nastala chyba jiz pri nacteni prvniho lexemu
        result = LEX_ERROR;
    else
        result = _prog();
    strFree(&attr);
    return result;
}

/*
TODO
        DONE
*/

/*  GRAMATIKA
+1	<prog>	 	-->	<declr> <stat_body> . <EOF>
+2	<declr>	 	-->	var <declr_list>
+3	<declr>	 	-->	begin
+4	<declr_list>	-->	"id : <type>; <is_declr_end>"
+5	<is_declr_end>	-->	<declr_list>
+6	<is_declr_end>	-->	begin
+7	<stat_body>	-->	end         // prazdny begin end.
+8	<stat_body>	-->	<stat_list>
+9	<type>	 	-->	string
+10	<type>	 	-->	integer
+11	<type>	 	-->	double
+12	<stat_list>	-->	<stat> <is_end>
+13	<is_end>	-->	"; <stat_list>"
+14	<is_end>	-->	end
+15	<stat>	 	-->	readln ( <id_list>
+16	<id_list>	-->	id <end_id_list>
+17	<end_id_list>	-->	, <id_list>
+18	<end_id_list>	-->	)       // prava zatvorka
+19	<stat>	 	-->	write ( <expr_list>
+20	<expr_list>	-->	<expr> <end_expr_list>
+21	<end_expr_list>	-->	, <expr_list>
+22	<end_expr_list>	-->	)       // prava zatvorka
+23	<stat>	 	-->	begin <is_empty>
+24	<is_empty>	-->	end          // prazdna podvetva
+25	<is_empty>	-->	<stat_list>
+26	<stat>	 	-->	if <expr> then <stat> <if_else>
+27	<if_else>	-->	else <stat>
+28	<if_else>	-->	epsilon
+29	<stat>	 	-->	while <expr> do <stat>
+30	<stat>	 	-->	id := <inner>
+31	<inner>	 	-->	<expr>
+32	<inner>	 	-->	sort ( expr )    // typova semantika!
+33	<inner>	 	-->	find ( expr , expr )

    // rozsirujuce pravidla 
+34	<stat>	 	-->	repeat <repeat_body> <expr>
+35	<repeat_body>   --> <stat> <repeat_end>
+36	<repeat_end>   --> ; <repeat_body>
+37	<repeat_end>   --> until
*/
