/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
/*      ###           Implementacia interpretu jazyka IFJ09          ###      */  
/* ########################################################################## */
/*  Akademicky rok: 2009/2010 ## Predmet: IFJ ## Datum odovzdania : 13.12 09  */
/* |------------------------------------------------------------------------| */
/* |        Team              |   Login   |       Nazov a popis modulu      | */ 
/* |------------------------------------------------------------------------| */
/* | Martin Haviar - veduci   | xhavia01  |   modul lexikalnej analyzi      | */
/* | Jozef  Henzl             | xhenzl04  |                                 | */
/* | Martin Herich            | xheric00  |                                 | */
/* | Maros  Barjak            | xbarja00  |                                 | */
/* \------------------------------------------------------------------------/ */

/* system includes */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdbool.h>
/* local includes */
#include "str.h"
#include "scanner.h"
#include "parser.h"

/**Jednotlive stavy konecneho automatu, pomocne aj koncove. Niektore jednoduche tokeny stavy ani nepotrebuju,ale
konecny automat obecne by mal mat tolko koncovych stavov kolko tokenov ma reprezentovat*/
typedef enum {

  START,                //zaciatocny stav z ktoreho vychadzame
  KOMENTAR,             //{ Komentar,pozor na koniec suboru pred ukoncovacou zatvorkou }
  EXTRA_KOMENTAR_1,     //rozsirenie,komentare ako su uvedene aj tu tzv "//komentar za dvoma slashmi"
  EXTRA_KOMENTAR_2,     //pomocny
  PLUS,                 //+
  MINUS,                //-
  NASOBENIE,            //*
  LAVA_ZATVORKA,        //(
  PRAVA_ZATVORKA,       //)
  STREDNIK,             //;
  CIARKA,               //,
  ROVNA_SA,             //=
  DOT_STATE,            //.
  DVOJBODKA,            //:
  PRIRADENIE,           //:=
  MENSIE,               //<
  MENSIE_ROVNE,         //<=
  NEROVNE,              //<>
  VACSIE,               //>
  VACSIE_ROVNE,         //>=
  ID_OR_KEYWORD,        // _mojaPremenna
  INTEGER_STATE,        // 543
  DOUBLE_DOT_STATE,     // 3.32
  DOUBLE_EXP_STATE,     //3e...
  DOUBLE_SIGN_STATE,    //3e+...
  DOUBLE_FINAL_DOT,     //23.432
  DOUBLE_FINAL_EXP,     //3.3e342
  BINARY_NUMBER_STATE,  //rozsirenie pre cisla v dvojkovej sustave,pomocny stav
  BINARY_NUMBER_OK,     //vysledny stav pre dvojkove cisla
  OCTA_NUMBER_STATE,    //cislo v osmickovej sustave,pomocny stav
  OCTA_NUMBER_OK,       //octa cislo v poriadku
  HEXA_NUMBER_STATE,    //hexadecimalne cislo,pomocny stav
  HEXA_NUMBER_OK,       //hexa cislo v poriadku
  STRING_STATE,         // 'retazec'
  END_STRING_CORRECTLY, //stav pre ukoncenie stringu

} Tstate;

/**globalna premenna pre jednoduchsiu pracu zo vstupnym suborom ,sem potom pojde argv[xy]*/
FILE *source=NULL;

/** radek na kterem prave jsme */
int lineno = 1;

/** 
@return lineno - riadok na ktorom sa nachadzame*/
int getCurLine(void) {

	return lineno;
}

/**funkcia pre otvorenie suboru
@param path cesta k suboru
@return RUN_ERROR pri chybnom otvoreni ,0 pri uspesnom otvoreni */
int ReadFile(char *path) {

    FILE *subor;
    if ((subor = fopen(path,"r")) == NULL)
        {          
          return RUN_ERROR;
        }
   	source = subor;       //nastavime
   	return 0;	// prebehlo v poriadku
}

/**Funkcia pre zatvorenie suboru,pokial niaky existuje*/
void CloseFile(void) {

	if(source != NULL)
	   fclose(source);
}

/**pri retazcoch mame escape sekvencie '#i'kde i patri do intervalu cisel <1,31>,preto robime kontrolu ci su cisla naozaj z tochto intervalu
@param buff cislo ktore kontrolujeme
@return 0 - uspech || 1 - neuspech  */
int numb_control(int buff) {
	return ((buff) >= 1 && (buff) <= 31  ) ? 0 : 1 ;
}


/**funkcia ktoru vola syntakticky analyzator ,funkcia vracia cislo ktore reprezentuje nadefinovany dany symbol,v parametri funkcie ,v bufferi vracia dany atribut / hodnotu symbolu ... napriklad ak nacitame identifikator ,funkcia vrati ze sa jedna o identifikator a v bufferi posle jeho nazov 
@param buffer buffer v ktorom funckia vrati atribut daneho tokenu ktory ho reprezentuje 
@return TYP TOKENU */
int getNextToken(string_t *buffer) {

 char int_buffer[3] = { 0 };   //pomocny buffer pre kontrolu rozsahu escape sekvencie i patri <1,31>,
 /// +1 index naviac kvoli interpretu a praci s pametous
 long int transfer;            //pomocna premenna pre konstanty v inych sustavach,strtoul vracia long int

Tstate state = START;       //pociatocni stav

int c=0;                     //nacitavany znak

strClear(buffer);         //vycistime buff
  while ((c = fgetc(source)) != EOF) 
   {//WHILE
         switch(state) {//SWITCH

             case START :
				if(c == '\n') {
					lineno++;
					state=START;
				} 
				else if (isspace(c))  //prazdne miesta a koniec riadkov preskocime,nemenime stav ...
                  state = START;
                //jednoduche stavy
                else if (c == '{') state = KOMENTAR;
                else if (c == '/') state = EXTRA_KOMENTAR_1;
                else if (c == '+') state = PLUS;
                else if (c == '-') state = MINUS;
                else if (c == '*') state = NASOBENIE;
                else if (c == '(') state = LAVA_ZATVORKA;
                else if (c == ')') state = PRAVA_ZATVORKA;
                else if (c == ';') state = STREDNIK;
                else if (c == ',') state = CIARKA;
                else if (c == '=') state = ROVNA_SA;
                else if (c == '.') state = DOT_STATE;
               //viacstavove
                else if (c == ':') state = DVOJBODKA;
                else if (c == '<') state = MENSIE;
                else if (c == '>') state = VACSIE;
                else if (c == '%') state = BINARY_NUMBER_STATE;   //rozsirenie,stav pre binarne cisla
                else if (c == '&') state = OCTA_NUMBER_STATE;     //cisla v osmickovej sustave
                else if (c == '$') state = HEXA_NUMBER_STATE;     //hexadecimalne cisla
                else if (isalpha(c) || c == '_')                  //identifikator alebo klucove slovo
                    {
                      strAddChar(buffer,c);    //ukladame po znakoch do bufferu
                      state = ID_OR_KEYWORD;   //nastavujeme pozadovane stavy
                    }
                else if (isdigit(c)) //cislo
                    {
                        strAddChar(buffer,c);
                        state = INTEGER_STATE;
                    }
                else if (c == '\'')   // retazec
                    {
                      state = STRING_STATE;
                    }
                else return WRONG_LEXEM;        //v pripade chybneho lexemu vraciame chybu
             break;

// ----------------------------------------------- JEDNODUCHE STAVY---------------------------------------------

             case KOMENTAR :
                if (c == '}') 
                   state = START;
                else 
                   state = KOMENTAR;   // vraciame sa stale do stavu komentar, nic dolezite pre nas ...
             break;

             case EXTRA_KOMENTAR_1 :
                   if (c == '/')   //spravne zapisany slahovy komentar
                      state = EXTRA_KOMENTAR_2;
                   else 
                      return WRONG_LEXEM;            //chybne zapisany,vraciame chybu
             break;

             case EXTRA_KOMENTAR_2 :
                  if (isprint(c) && c != '\n') //komentar pokracuje do konca riadka
                      state = EXTRA_KOMENTAR_2;  //ostavam vnom
                  else if (c == '\n') {
		      lineno++;           //pocitam riadky aby som vedel kde sa nachadzam
                      state = START;
				  }
             break; 

             case PLUS :          //pri nacitani dalsieho znaku ten aktualny vratime a prislusny vtomto pripade 
                ungetc(c,source);  //PLUS_TOKEN vratime
                return PLUS_TOKEN;
             break;
  
             case MINUS :
                ungetc(c,source);
                return MINUS_TOKEN;
             break;

             case NASOBENIE :
                ungetc(c,source);
                return MULT_TOKEN;
             break;

             case LAVA_ZATVORKA :
                ungetc(c,source);
                return LEFT_BRACKET;
             break;

             case PRAVA_ZATVORKA :
                ungetc(c,source);
                return RIGHT_BRACKET;
             break;

             case STREDNIK :
                ungetc(c,source);
                return SEMICOLON;
             break;

             case CIARKA :
                ungetc(c,source);
                return COMMA;
             break;

             case ROVNA_SA :
                ungetc(c,source);
                return EKVIVALENCE;
             break;

             case DOT_STATE :
                 ungetc(c,source);
                 return DOT_TOKEN;
             break;

//------------------------------------------ZLOZITE STAVY-----------------------------------------------------------

             case DVOJBODKA :
               if (c == '=')
                 state = PRIRADENIE;
               else {
                  ungetc(c,source);
                  return COLON;
                }
             break;

             case PRIRADENIE :
                 ungetc(c,source);
                 return ASSIGNING;
             break;
             //vacsie mensie a tak
             case MENSIE :
               if (c == '=') 
                   state = MENSIE_ROVNE;
               else if (c == '>')
                   state = NEROVNE;
               else {
                 ungetc(c,source);
                 return LESS_TOKEN;
               }
             break;

             case MENSIE_ROVNE :
                ungetc(c,source);
                return LESS_EQUAL;
             break;

             case NEROVNE :
                ungetc(c,source);
                return UN_EQUAL;
             break;

             case VACSIE :
               if (c == '=') 
                  state = VACSIE_ROVNE;
               else {
                  ungetc(c,source);
                  return GREATER_TOKEN;
               }
             break;

             case VACSIE_ROVNE :
                ungetc(c,source);
                return GREATER_EQUAL;
             break;
             // ID    ++++   KEYWORDS
             case ID_OR_KEYWORD :
                if (isalnum(c)) {    //ID pokracuje cislami alebo pismenami
                   strAddChar(buffer,c); 
                   state = ID_OR_KEYWORD;
                }  
                else { //koniec ID ,porovname ci nieje klucove slovo
                  ungetc(c,source);
                  
                  if (strCmpConstStr(buffer,"begin") == 0) return BEGIN;
                  else if (strCmpConstStr(buffer,"div") == 0) return DIV;
                  else if (strCmpConstStr(buffer,"do") == 0) return DO;
                  else if (strCmpConstStr(buffer,"double") == 0) return DOUBLE;
                  else if (strCmpConstStr(buffer,"else") == 0) return ELSE;
                  else if (strCmpConstStr(buffer,"end") == 0) return END;
                  else if (strCmpConstStr(buffer,"find") == 0) return FIND;
                  else if (strCmpConstStr(buffer,"if") == 0) return IF;
                  else if (strCmpConstStr(buffer,"integer") == 0) return INTEGER;
                  else if (strCmpConstStr(buffer,"readln") == 0) return READLN;
                  else if (strCmpConstStr(buffer,"sort") == 0) return SORT;
                  else if (strCmpConstStr(buffer,"string") == 0) return STRING;
                  else if (strCmpConstStr(buffer,"then") == 0) return THEN;
                  else if (strCmpConstStr(buffer,"var") == 0) return VAR;
                  else if (strCmpConstStr(buffer,"while") == 0) return WHILE;
                  else if (strCmpConstStr(buffer,"write") == 0) return WRITE;
                  else if (strCmpConstStr(buffer,"repeat") == 0) return REPEAT;
                  else if (strCmpConstStr(buffer,"until") == 0) return UNTIL;
                  else
                   return ID;   //ak neni klucove slovo vrati ze sa jedna o identifikator

                }
              break;

           // CISELNE KONSTANTY
              case INTEGER_STATE :
                   if (isdigit(c)) {     //ostavame v inte
                     strAddChar(buffer,c);
                     state = INTEGER_STATE;
                   }
                   else if (c == '.') {    //priznak double cisla
                     strAddChar(buffer,c);
                     state = DOUBLE_DOT_STATE;
                   }
                   else if (c == 'e' || c == 'E') {  //double cislo 
                     strAddChar(buffer,c);
                     state = DOUBLE_EXP_STATE;
                   }
                   else {
                      ungetc(c,source);
                      return INT_CONST;
                   }
              break;

              case DOUBLE_DOT_STATE :
                    if (isdigit(c)) {
                       strAddChar(buffer,c);
                       state = DOUBLE_FINAL_DOT;
                    }
                    else  {
                       strAddChar(buffer,c);
                       return WRONG_LEXEM;
                    }
              break;

              case DOUBLE_EXP_STATE :
                    if (c == '+' || c == '-') {
                       strAddChar(buffer,c);
                       state = DOUBLE_SIGN_STATE;
                    }
                    else if (isdigit(c)) {
                       strAddChar(buffer,c);
                       state = DOUBLE_FINAL_EXP;
                    }
                 else  {
                    strAddChar(buffer,c);
                    return WRONG_LEXEM;
                    }
              break;

              case DOUBLE_FINAL_DOT :
                 if (isdigit(c)) {
                     strAddChar(buffer,c); 
                     state = DOUBLE_FINAL_DOT;
                 }
                 else if (c == 'e' || c == 'E') {
                     strAddChar(buffer,c);
                     state = DOUBLE_EXP_STATE;
                 }
                 else {
                      ungetc(c,source);
                      return DOUBLE_DOT;
                 }
              break;
 
              case DOUBLE_SIGN_STATE :
                   if (isdigit(c)) {
                     strAddChar(buffer,c);  
                     state = DOUBLE_FINAL_EXP;
                   }
                   else  {
                     strAddChar(buffer,c);
                     return WRONG_LEXEM;
                    }
              break;

              case DOUBLE_FINAL_EXP :
                   if (isdigit(c)) {
                     strAddChar(buffer,c);
                     state = DOUBLE_FINAL_EXP;
                   }
                   else {
                      ungetc(c,source);
                      return DOUBLE_EXP;
                   }
              break; 

              case BINARY_NUMBER_STATE :        //%01010111101100111001
                   if (c == '0' || c == '1') {  //moze nasledovat len 0 alebo 1
                       strAddChar(buffer,c);
                       state = BINARY_NUMBER_OK;
                       }
                   else { //%3  napr je chybne cislo teda znak % vratime ako W_L a trojka je normalny dalsi lexem
                       ungetc(c,source);
                       return WRONG_LEXEM;
                   }
              break;

              case BINARY_NUMBER_OK :
                   if (c == '0' || c == '1') {
                     strAddChar(buffer,c);
                     state = BINARY_NUMBER_OK;   //nacitame cele binarne cislo
                   }
                   else {
                     ungetc(c,source);   //po ukonceni binarneho cisla
                     //ho prevedieme z dvojkovej sustavy do desiatkovej a vlozime do pomocnej premennej
                     transfer = strtol(buffer->str,NULL,2);
                     strClear(buffer);  //vyprazdnime buffer
                     sprintf(buffer->str,"%ld", transfer);  //a zapiseme donho cislo v desiatkovej sustave
                     return INT_CONST;   //vratime cislo 
                   }
              break;

              case OCTA_NUMBER_STATE : //osmickove cislo je z cislic 0 az 7
                   if (c >= '0' && c <= '7') {
                     strAddChar(buffer,c);
                     state = OCTA_NUMBER_OK;
                   }
                   else {
                     ungetc(c,source);
                     return WRONG_LEXEM;
                   }

              break;

              case OCTA_NUMBER_OK :
                   if (c >= '0' && c <= '7') {
                     strAddChar(buffer,c);
                     state = OCTA_NUMBER_OK;
                   }
                   else {
                     ungetc(c,source);
                     transfer = strtol(buffer->str,NULL,8);
                     strClear(buffer);
                     sprintf(buffer->str,"%ld", transfer);
                     return INT_CONST;
                   }
              break;

              case HEXA_NUMBER_STATE :    //hexa cislo
                   if (isxdigit(c)) {
                     strAddChar(buffer,c);
                     state = HEXA_NUMBER_OK;
                   }
                   else {
                     ungetc(c,source);
                     return WRONG_LEXEM;
                   }
              break;

              case HEXA_NUMBER_OK :
                   if (isxdigit(c)) {
                    strAddChar(buffer,c);
                    state = HEXA_NUMBER_OK;
                   }
                   else {
                     ungetc(c,source);
                     transfer = strtol(buffer->str,NULL,16);
                     strClear(buffer);
                     sprintf(buffer->str,"%ld", transfer);
                     return INT_CONST;
                   }
              break;

              // RETAZCE 
              case STRING_STATE :
                  if (c > 31 && c != '\'') {    //v retazci ostavame,v stringu mozu byt mimo escape sek. len znaky s asci hodnotou > ako 31
                      strAddChar(buffer,c);
                      state = STRING_STATE;
                  }
                  else if (c <= 31) 
                      return WRONG_LEXEM; 

                  else if (c == '\'') {       //dvojity apostrof alebo escape sekvencia alebo koniec retazca
                      c = fgetc(source);      //nacitam si dalsi znak dopredu aby som vedel co ide
                        if (c == '#')         // ide escape sekvencia
                      {  
                               int_buffer[0] = 0;
                               int_buffer[1] = 0;
                               int_buffer[2] = 0;
                          do  { c = fgetc(source);  //
                          } while(c=='0');   //nuly sa v esc.sekvenci mozu vystkytnut preto ich preskocim
                            if ((!(isdigit(c)))){ //ak sa nejedna o cislo
                                 ungetc(c,source);    //vratime znak
				 return WRONG_LEXEM;
                            } 
                             int cislo = 0;     //pomocna premenna pre kontrolu cisla -> rozsach 'i' pri esc sekvencii
                             int_buffer[0] = c;  //vlozime prvy znak(cislo ako char)
                             cislo = atoi(int_buffer); //priradim prve cislo
                             c = fgetc(source);  //nacitame dalsi
                            if (isdigit(c)){  //ak je cislo
                             int_buffer[1] = c;  //ukladame druhe
                             cislo = atoi(int_buffer); //do pomocnej si prevedieme charove cisla na integer
                            }
                            else if (c != '\'')
                               return WRONG_LEXEM; //ak druhe cislo neni cislo
                            if (numb_control(cislo) != 0){     //kontrola rozsahu,ak je mimo nasleduje chybovy stav
                             ungetc(c,source);
			     return WRONG_LEXEM;
                            }
                            if (c != '\'') c = fgetc(source);   //cislo moze byt aj jednociferne !toto je preto
                            if (c=='\'') {    //spravne ukoncenie escape sekvencie ,vraciame sa do string_state-u
                             state = STRING_STATE;
                             strAddChar(buffer,cislo);
                            }
                            else {    //nespravna escape sekvencia ,chyba ...
                             ungetc(c,source);
                   	     return WRONG_LEXEM;
                            }
                       }//tuto dole pri spravnom ukonceni stringu mozu byt prave tieto znaky
                      else if (isspace(c) || c == '+' || c == ',' || c == ';' || c == ')' || c == '<' ||
                                             c == '>' || c == '-' || c == '=' || c=='(' || c==':'){  

                              ungetc(c,source);//navratime
                              state = END_STRING_CORRECTLY;  //ideme kontrolovat koniec retazca

                      }
                      else if (c == '\''){     //double apostrophe
                              strAddChar(buffer,c);  //ulozime,tento sa ma zobrazit
                              state = STRING_STATE;
                              break;
                      }
                      else {     //inac chyba
                           ungetc(c,source);
                           return WRONG_LEXEM;
                       }
                  }
              break; 

              case END_STRING_CORRECTLY :  //za stringom sa mozu nachadzat len operatory nasej GRAMATIKY
                    //syntakticku chybu kontrole syntakticky analyzator !!!
                    if (isspace(c) || c == '+' || c == ',' || c == ';' || c == ')' || c =='-' || c == '(' ||
                                      c == '=' || c==':' || c =='<' || c =='>') {
                       ungetc(c,source);//vratime ho a
                       return STRING;  //return spravny string

                    }
                    else { //inak lexikalna chyba
                     return WRONG_LEXEM;
                    }
              break;

        } //END SWITCH

   } //END WHILE

if (state == KOMENTAR) {  //ak sme ostali v stave komentar a subor uz vyskocil z hlavneho cyklu znamena to ze komentar
   return WRONG_LEXEM;   //nebol ukonceny a dojde ku chybe
 }

if (c==EOF) { //ak je koniec suboru
   if (state == STRING_STATE)//ak dojde ku koncu suboru v string_state-e
       return WRONG_LEXEM;  //vraciame chybu
   else
       return END_OF_FILE;   //ak nie klasicky vraciame koniec suboru 
 }

strFree(buffer); //uvolnime buffer
return 0;
}

#if 0

int main(void) {


/////////////////////////////////////////
ReadFile("test.txt");   //testovaci subor
////////////////////////////////////////////


int result;
string_t *retazec;

retazec = (string_t*) malloc(sizeof(string_t)); //v SYNTAKTICKEJ KTORA TO BUDE SPRACUVAT TREBA TAKTO NAALOKOVAT
strInit(retazec);      //RETAZEC TREBA INICIALIZOVAT


result = getNextToken(retazec);
while (result !=END_OF_FILE)
{
	printf("%s\t%d\n",retazec->str,result);
	result = getNextToken(retazec);
}
printf("%s\t%d\n",retazec->str,result); // vypis EOF

/////////////////
CloseFile();
////////////////

strFree(retazec);
free(retazec);


return 0;
}
#endif

