/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
/*      ###           Implementacia interpretu jazyka IFJ09          ###      */  
/* ########################################################################## */
/*  Akademicky rok: 2009/2010 ## Predmet: IFJ ## Datum odovzdania : 13.12 09  */
/* |------------------------------------------------------------------------| */
/* |        Team              |   Login   |       Nazov a popis modulu      | */ 
/* |------------------------------------------------------------------------| */
/* | Martin Haviar - veduci   | xhavia01  |   modul pre pracu s retazcami   | */
/* | Jozef  Henzl             | xhenzl04  |                                 | */
/* | Martin Herich            | xheric00  |                                 | */
/* | Maros  Barjak            | xbarja00  |                                 | */
/* \------------------------------------------------------------------------/ */
                                                                 

                                                                 

//jednoducha knihovna pro praci s nekonecne dlouhymi retezci
#include <string.h>
#include <malloc.h>
#include "str.h"

#define STR_LEN_INC 8
// konstanta STR_LEN_INC udava, na kolik bytu provedeme pocatecni alokaci pameti
// pokud nacitame retezec znak po znaku, pamet se postupne bude alkokovat na
// nasobky tohoto cisla 

#ifndef strdup
/* vlastni strdup */
char *strdup (const char *s)
{
	char *ret;
	ret = malloc(strlen(s)+1);
	if(ret == NULL) return NULL;
	strcpy(ret, s); 
	return ret;
}
#endif


int strInit(string_t *s)
// funkce vytvori novy retezec
{
   if ((s->str = (char*) malloc(STR_LEN_INC)) == NULL)
      return STR_ERROR;
   s->str[0] = '\0';
   s->length = 0;
   s->allocSize = STR_LEN_INC;
   return STR_SUCCESS;
}

void strFree(string_t *s)
// funkce uvolni retezec z pameti
{
   strClear(s); 
   free(s->str);
   s->str = NULL;
}

void strClear(string_t *s)
// funkce vymaze obsah retezce
{
   s->str[0] = '\0';
   s->length = 0;
}

int strAddChar(string_t *s1, char c)
// prida na konec retezce jeden znak
{
   if (s1->length + 1 >= s1->allocSize)
   {
      // pamet nestaci, je potreba provest realokaci
      if ((s1->str = (char*) realloc(s1->str, s1->length + STR_LEN_INC)) == NULL)
         return STR_ERROR;
      s1->allocSize = s1->length + STR_LEN_INC;
   }
   s1->str[s1->length] = c;
   s1->length++;
   s1->str[s1->length] = '\0';
   return STR_SUCCESS;
}

int strCmpString(string_t *s1, string_t *s2)
// porovna oba retezce a vrati vysledek
{
   return strcmp(s1->str, s2->str);
}

int strCmpConstStr(string_t *s1, char* s2)
// porovna nas retezec s konstantnim retezcem
{
   return strcmp(s1->str, s2);
}

/**
* spoji retezec s C null-terminated stringem
*/
int strConcat(string_t *s1, char *cs)
{
	int cs_len;
	int alloc=0;

	cs_len = strlen(cs);

	if(cs_len == 0) return STR_SUCCESS;

	if ((s1->length + cs_len) >= s1->allocSize)
	{
		// pamet nestaci, je potreba provest realokaci
		if(cs_len < STR_LEN_INC) 
			alloc = STR_LEN_INC;
		else
			alloc = cs_len;
		if ((s1->str = (char*) realloc(s1->str, s1->allocSize + alloc)) == NULL)
			return STR_ERROR;
		s1->allocSize = s1->allocSize + alloc;
	}
	strcat(s1->str, cs);
	s1->length += cs_len ;
	s1->str[s1->length] = '\0';
	return STR_SUCCESS;
}

int strDup(string_t *dst, string_t *src)
{
	strClear(dst);
	strConcat(dst, src->str);
	if(dst->str == NULL) return -1;
	return 0;
}

/** funkce find */
#define K  255
/* shiftovaci vektor */
static int sh[K];

/**
* inicializace shiftovaciho vektoru - v tabulce nastavime index znaku, ktery
	se ve vzorku nachazi, na vzdalenost jeho prvniho vyskytu ve vzorku zprava
*/
static void __init_shift(char *str)
{
	int i;
	int str_len  = strlen(str);
	for(i=0;i<K;i++) sh[i] = str_len;
	for(i=0;i<str_len;i++) sh[(int)(str[i])] = str_len-i-1;
}

/**
* samotna vyhledavaci funkce
*/
int find(char *str, char *tok)
{
	int i, j, tok_len, str_len;
	int x;

	tok_len = strlen(tok);
	str_len = strlen(str);

	__init_shift(tok);

	for(i = tok_len - 1, j = tok_len - 1; j >= 0; i--, j--)
		while(tok[j] != str[i]) {
			/* pocet shiftnuti */
			x = sh[(int)(str[i])];
			if((tok_len - j) > x)
				i += tok_len - j;
			else i += x;
			if(i >= str_len) return -1;
			j = tok_len - 1;
		}
	return i+1;
}
#if 0
int main(int argc, char **argv)
{
	char *str = "lorem ipsum dolor amet";
	int res;

	char *tok;
	if(argc < 3) {
		fprintf(stderr, "%s <tok> <string>\n", argv[0]);
		return -1;
	}
	str = argv[2];
	tok = argv[1];

	res = find(str, tok);

	printf("%d\n", res);
	return 0;
}
#endif
/* konec find */

#if 0
#include <stdio.h>
int main(int argc, char **argv)
{

	string_t new;
	
	strInit(&new);
	
	strConcat(&new, "kocka");
	strConcat(&new, " leze");
	strConcat(&new, " dirou,\n");
	strConcat(&new, "pes");
	strConcat(&new, " oknem,\n");
	strConcat(&new, "nebude-li");
	strConcat(&new, " prset,\n");
	strConcat(&new, "nezmoknem\n");
	
	printf("%s", new.str);

	strFree(&new);
}
#endif


/*************** SORT  ***************/
#define SORT_DEBUG 0

/* auxiliary array for merge function */
static char *_aux_arr=NULL;
static int _aux_arr_len=0;

static int __sort(char *str, int lo, int hi);
static int __merge(char *str, int lo, int mid, int hi);

int sort(char *str)
{
	int ret;
	_aux_arr_len = strlen(str);
	_aux_arr = malloc(sizeof(char) * _aux_arr_len);
	if(_aux_arr == NULL) return -1;
	ret = __sort(str, 0, _aux_arr_len-1);
	free(_aux_arr);
	return ret; 
}

static int __sort(char *str, int lo, int hi)
{
	int mid;
	if(lo < hi) {
		mid = (lo+hi)/2;
		__sort(str, lo, mid);
		__sort(str, mid+1, hi);
		return __merge(str, lo, mid, hi);
	}
	return 0;
}


/**
* merge two sorted parts
* @param str memory with sorted parts
* @param lo index of first item in array
* @param mid end of first part
* @param hi end of second part
*/
static int __merge(char *str, int lo, int mid, int hi)
{
	int i, j, k;

	memset(_aux_arr, 0, _aux_arr_len);	

#if SORT_DEBUG
	for (i=lo;i<=hi;i++)
		printf("%c", str[i]);
	printf("->");
#endif
	/* copy both parts to aux array */
	for(i = lo; i <= hi; i++) _aux_arr[i] = str[i];

	i = k = lo;
	j = mid + 1;
	/* copy greatest element back to str */
	while((i <= mid) && (j <= hi)) {
		if(_aux_arr[i] <= _aux_arr[j])	str[k++] = _aux_arr[i++];
		else 					str[k++] = _aux_arr[j++];
	}

	/* copy back the rest */
	while(i<= mid) str[k++] = _aux_arr[i++];

#if SORT_DEBUG
	for (i=lo;i<=hi;i++)
		printf("%c", str[i]);

	putchar('\n');
#endif

	return 0;	
}

#if SORT_DEBUG


int main(int argc, char **argv)
{
	char *mystr;

	mystr = strdup(argv[1]);

	printf("%s\n", mystr);
	sort(mystr);
	printf("%s\n", mystr);

	free(mystr);
	return 0;
}
#endif
/***************** /SORT	*****************************************/
